/*****************************************************************************
 *
 *  $Id: pdserv.h,v fa0afd9f0a9e 2015/12/21 10:02:02 lerichi $
 *  vim:ft=c:tw=78
 *
 *  Copyright 2010 - 2016  Richard Hacker (lerichi at gmx dot net)
 *
 *  This file is part of the pdserv library.
 *
 *  The pdserv library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  The pdserv library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with the pdserv library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

/** @file
 * @example example-st.c
 */

#ifndef PDSERV_H
#define PDSERV_H

/** \cond */

#include <stddef.h>


#ifdef pdserv_EXPORTS
#  define PDSERV_EXPORT __attribute__((__visibility__("default")))
#else
#  define PDSERV_EXPORT
#endif


/** \endcond */

/** \def PDSERV_VERSION
 * \brief Generate version code from major, minor and patchlevel
 *
 * \def PDSERV_VERSION_CODE
 * \brief Current library version code
 *
 * \def PDSERV_VERSION_MAJOR
 * \brief Current library major version code
 *
 * \def PDSERV_VERSION_MINOR
 * \brief Current library minor version code
 *
 * \def PDSERV_VERSION_PATCH
 * \brief Current library patch version code. Note: this may be a string!
 */
#define PDSERV_VERSION(major,minor,patchlevel) \
    (((major) << 16) + ((minor) << 8) + (patchlevel))
#define PDSERV_VERSION_CODE \
    PDSERV_VERSION(3,3,1)
#define PDSERV_VERSION_MAJOR 3
#define PDSERV_VERSION_MINOR 3
#define PDSERV_VERSION_PATCH 1

#ifdef __cplusplus
extern "C" {
#endif

/**
 * \def HAS_VERSION_CODE
 * \brief Macro to indicate that there is a pdserv_version_str
 *
 * \var pdserv_version_str
 * \brief String of pdserv version code "major.minor.patch"
 * \var pdserv_full_version
 * \brief Full String of pdserv version, generated by <tt>git describe</tt>.
 */
#define HAS_VERSION_CODE
PDSERV_EXPORT extern const char* const pdserv_version_str;
PDSERV_EXPORT extern const char* const pdserv_full_version;

/* Data type definitions. */
/**
 * \def pd_double_T
 * \brief double precision floating point data type
 *
 * \def pd_single_T
 * \brief single precision floating point data type
 *
 * \def pd_uint8_T
 * \brief 8 bit unsigned integer data type
 *
 * \def pd_uint16_T
 * \brief 16 bit unsigned integer data type
 *
 * \def pd_uint32_T
 * \brief 32 bit unsigned integer data type
 *
 * \def pd_uint64_T
 * \brief 64 bit unsigned integer data type
 *
 * \def pd_sint8_T
 * \brief 8 bit signed integer data type
 *
 * \def pd_sint16_T
 * \brief 16 bit signed integer data type
 *
 * \def pd_sint32_T
 * \brief 32 bit signed integer data type
 *
 * \def pd_sint64_T
 * \brief 64 bit signed integer data type
 *
 * \def pd_boolean_T
 * \brief boolean data type
 *
 * \def pd_schar_T
 * \brief signed char data type
 *
 * \def pd_char_T
 * \brief char data type
 *
 * \def pd_uchar_T
 * \brief unsigned char data type
 *
 * \def pd_short_T
 * \brief short data type
 *
 * \def pd_ushort_T
 * \brief unsigned short data type
 *
 * \def pd_int_T
 * \brief int data type
 *
 * \def pd_uint_T
 * \brief unsigned int data type
 *
 * \def pd_long_T
 * \brief long data type
 *
 * \def pd_ulong_T
 * \brief unsigned long data type
 *
 * \def pd_longlong_T
 * \brief long long data type
 *
 * \def pd_ulonglong_T
 * \brief unsigned long long data type
 *
 * \def pd_ssize_T
 * \brief signed size_t data type
 *
 * \def pd_size_T
 * \brief size_t data type
 *
 */
/* Let the enumeration start at 1 so that an unset data type could be
 * detected.*/
#define pd_double_T      1
#define pd_single_T      2
#define pd_uint8_T       3
#define pd_sint8_T       4
#define pd_uint16_T      5
#define pd_sint16_T      6
#define pd_uint32_T      7
#define pd_sint32_T      8
#define pd_uint64_T      9
#define pd_sint64_T     10
#define pd_boolean_T    11
#define pd_schar_T      12
#define pd_char_T       13
#define pd_uchar_T      14
#define pd_short_T      15
#define pd_ushort_T     16
#define pd_int_T        17
#define pd_uint_T       18
#define pd_long_T       19
#define pd_ulong_T      20
#define pd_longlong_T   21
#define pd_ulonglong_T  22
#define pd_ssize_T      23
#define pd_size_T       24
/** \cond Doxygen must ignore pd_datatype_end */
#define pd_datatype_end 25

struct timespec;

/** \endcond */

/* Structure declarations.
 */
struct pdserv;          /**< Generic handle to global library instance */
struct pdtask;          /**< Generic handle to a task */
struct pdvariable;      /**< Generic handle to a variable */
struct pdevent;         /**< Generic handle to an event */


/** Callback function type to get current world time */
typedef int (*gettime_t)(struct timespec*);

/** Initialise pdserv library.
 *
 * This is the first call to initialize an instance of the library. It must be
 * called before any other library calls.
 *
 * \return
 *      - handle to an opaque instance of pdserv
 *      - NULL on error
 */
PDSERV_EXPORT struct pdserv* pdserv_create(
        const char *name,       /**< Name of the process */
        const char *version,    /**< Version string */
        gettime_t gettime_cb    /**< Function used by the library
                                 * when the system time is required.
                                 * Essentially, this function should call
                                 * clock_gettime().
                                 * If NULL, gettimeofday() is used. */
        );

/** Set the path to a configuration file
 *
 * This file will loaded during pdserv_prepare() in
 * non-real-time context.
 *
 * Failure to load will be reported in syslog.
 *
 * If a configuration file is not specified, it will attempt to load
 *    - PDSERV_CONFIG environment variable
 *    - /etc/pdserv.conf
 */
PDSERV_EXPORT void pdserv_config_file(
        struct pdserv* pdserv,  /**< Handle to pdserv */
        const char *file        /**< Path to configuration file */
        );

/** Register a function to be called when setting a parameter's value.
 *
 * Since parameter writes from a client application are asynchronous as far as
 * the real-time threads are concerned, access to them must be protected using
 * a lock or inconsistent data is risked.
 *
 * This function registers a callback which is called before (with \b
 * lock = 1) and after (with \b lock = 0) setting the parameter value. Inside
 * the callback a lock can be aquired and realeased as required, for instance.
 */
PDSERV_EXPORT void pdserv_set_parameter_writelock_cb(
        struct pdserv* pdserv,  /**< Handle to pdserv */
        void (*fn)(int lock, void* priv_data),/**< function to call */
        void* priv_data         /**< private data passed to callback */
        );

/** Create a cyclic task
 *
 * An application consists of one or more tasks that are called cyclically by
 * caller. Tasks are used to organize signals.
 *
 * \return
 *      - handle to an opaque task instance
 *      - NULL on error
 */
PDSERV_EXPORT struct pdtask* pdserv_create_task(
        struct pdserv* pdserv,  /**< Handle to pdserv */
        double tsample,         /**< Base sample time of task */
        const char *name        /**< Optional name; can be NULL */
        );

/** Register a function to be called when polling a signal's value.
 *
 * Since polling a signal value from a client application is asynchronous as
 * far as the real-time thread is concerned, access to it must be protected
 * using a lock or inconsistent data is risked.
 *
 * This function registers a callback which is called before (with \b
 * lock = 1) and after (with \b lock = 0) copying the signal value. Inside the
 * callback a lock can be aquired and realeased as required, for instance.
 */
PDSERV_EXPORT void pdserv_set_signal_readlock_cb(
        struct pdtask* pdtask,  /**< Handle to pdtask */
        void (*fn)(int lock, void* priv_data),/**< function to call */
        void* priv_data         /**< private data passed to callback */
        );

/** Create compound data type
 *
 * Use this to create an application specific data type.
 * Use pdserv_compound_add_field() to insert fields
 *
 * \return
 *      compound data type identifier
 */
PDSERV_EXPORT int pdserv_create_compound(
        const char *name,       /**< Data type name */
        size_t size             /**< Size of compound */
        );

/** Add a field to a compound data type
 */
PDSERV_EXPORT void pdserv_compound_add_field(
        int compound,           /**< Compound data type identifier */
        const char *name,       /**< Data type name */
        int data_type,          /**< Field data type identifier:
                                 * - pd_*_T
                                 * - compound identifier from
                                 *   pdserv_create_compound() */
        size_t offset,          /**< Offset of field; use offsetof() */
        size_t ndim,            /**< Number of dimensions in \p dim */
        const size_t *dim       /**< Dimensions */
        );

/** Register a signal
 *
 * This call registers a signal, i.e. Variables that are calculated during
 * a task's execution.
 *
 * Arguments \p n and \p dim are used to specify the shape:
 *  - For scalars:
 *    * n: 1
 *    * dim: NULL
 *
 *  - For vectors length x
 *    * n: x
 *    * dim: NULL
 *
 *  - For arrays:
 *    \code{.c}
 *    int array[2][3][4]
 *    size_t array_dim[] = {2,3,4};
 *    \endcode
 *
 *    * n: 3
 *    * dim: array_dim
 *
 * Argument \p decimation specifies calculation decimation of the signal within
 * the task if it is not calculated every time.
 */
PDSERV_EXPORT struct pdvariable *pdserv_signal(
        struct pdtask* pdtask,    /**< Handle to pdtask */
        unsigned int decimation,  /**< Decimation with which the signal is
                                   * calculated */
        const char *path,       /**< Signal path */
        int datatype,           /**< Signal data type identifier
                                 * - pd_*_T
                                 * - compound identifier from
                                 *   pdserv_create_compound() */
        const void *addr,       /**< Signal address */
        size_t n,               /**< Element count.
                                 * If \p dim != NULL, this is the number
                                 * elements in \p dim */
        const size_t *dim       /**< Dimensions. If NULL, consider the
                                 * parameter to be a vector of length \p n */
        );

/** \cond
 * These functions are not required any more. They were used to perform
 * locking, but that is achieved using pdserv_set_signal_readlock_cb() */
typedef int (*read_signal_t)(const struct pdvariable *signal, void *dst,
        const void *src, size_t len, struct timespec* time, void *priv_data);

PDSERV_EXPORT void pdserv_signal_set_read_cb(struct pdvariable* signal,
        read_signal_t read_signal_cb, void* priv_data
        ) __attribute__((deprecated("use pdserv_signal() instead")));

PDSERV_EXPORT struct pdvariable *pdserv_signal_cb( struct pdtask* pdtask,
        unsigned int decimation, const char *path, int datatype,
        const void *addr, size_t n, const size_t *dim,
        read_signal_t read_signal_cb, void* priv_data
        ) __attribute__((deprecated("use pdserv_signal() instead")));;
/** \endcond */

/**
 * \def EMERG_EVENT
 * \brief Emergency event
 *
 * \def ALERT_EVENT
 * \brief Alert event
 *
 * \def CRIT_EVENT
 * \brief Critical event
 *
 * \def ERROR_EVENT
 * \brief Error event
 *
 * \def WARN_EVENT
 * \brief Warning event
 *
 * \def NOTICE_EVENT
 * \brief Notice event
 *
 * \def INFO_EVENT
 * \brief Information event
 *
 * \def DEBUG_EVENT
 * \brief Debug event
 *
 * \def RESET_EVENT
 * \brief Reset the event
 */
#define RESET_EVENT     0
#define EMERG_EVENT     1
#define ALERT_EVENT     2
#define CRIT_EVENT      3
#define ERROR_EVENT     4
#define WARN_EVENT      5
#define NOTICE_EVENT    6
#define INFO_EVENT      7
#define DEBUG_EVENT     8

/** Register an event channel
 *
 * Event channels are used to implement a messaging system. A clear text
 * message corresponding to the event element is printed when the event is set
 * using pdserv_event_set().
 *
 */
PDSERV_EXPORT struct pdevent *pdserv_event(
        struct pdserv* pdserv,  /**< Handle to pdserv */
        const char *path,       /**< Event path */
        size_t n                /**< Vector size */
        );

/** Set a text message for events
 */
PDSERV_EXPORT void pdserv_event_set_text(
        struct pdevent* event,  /**< Handle to event */
        const char * const *text /**< Event text list, same size as event
                                 *
                                 * PdServ does not make a copy of these
                                 * messages. They must be available
                                 * permanently. */
        );


/** Set the state of a single event.
 *
 * This function is used to set an event to the given state. Setting the event
 * causes the corresponding message to be printed to log and all clients.
 */
PDSERV_EXPORT void pdserv_event_set(
        const struct pdevent *event,  /**< Opage event structure */
        size_t element,         /**< Event element */
        int priority,           /**< One of:
                                 *   * #EMERG_EVENT
                                 *   * #ALERT_EVENT
                                 *   * #CRIT_EVENT
                                 *   * #ERROR_EVENT
                                 *   * #WARN_EVENT
                                 *   * #NOTICE_EVENT
                                 *   * #INFO_EVENT
                                 *   * #DEBUG_EVENT
                                 */
        const struct timespec *t /**< Current time */
        );

/** Reset an event.
 *
 * This function is used to reset an event. This causes the corresponding
 * message to be printed to log and all clients.
 */
PDSERV_EXPORT void pdserv_event_reset(
        const struct pdevent *event,  /**< Opage event structure */
        size_t element,         /**< Event element */
        const struct timespec *t /**< Current time */
        );

/** Set the state of event vector.
 *
 * This function is used to set an event to the given state. Setting the event
 * causes the corresponding message to be printed to log and all clients.
 *
 * This call is similar to pdserv_event_set(), except that its state state is
 * stored internally.  Only when \p level changes will the event be generated
 * or reset. Thus this function can be called continuously from the cyclic
 * task.
 *
 * Do not mix the calls to pdserv_event_set_state()
 * and pdserv_event_set() pdserv_event_reset() pair
 */
PDSERV_EXPORT void pdserv_event_set_all(
        const struct pdevent *event,  /**< Opage event structure */
        const unsigned int * level,  /**< One of:
                                 *   * #EMERG_EVENT
                                 *   * #ALERT_EVENT
                                 *   * #CRIT_EVENT
                                 *   * #ERROR_EVENT
                                 *   * #WARN_EVENT
                                 *   * #NOTICE_EVENT
                                 *   * #INFO_EVENT
                                 *   * #DEBUG_EVENT
                                 *   * #RESET_EVENT
                                 *
                                 *   The vector length must be the same as
                                 *   that specified when calling pdserv_event()
                                 */
        const struct timespec *t /**< Current time */
        );

/** Callback on a parameter update event
 *
 * This function is responsible for copying the data from \p src to \p dst.
 *
 * The function allows the user to synchronize his application to
 * the external events of setting a parameter.
 *
 * Essentially the function should execute the following:
 * \code{.c}
 *      lock(priv_data);
 *      memcpy(dst, src, len);
 *      clock_gettime(CLOCK_REALTIME, time);
 *      unlock(priv_data);
 * \endcode
 *
 * @param param Handle to parameter
 * @param dst   Copy destination address
 * @param src   Copy source address
 * @param len   Bytes to copy
 * @param time  Copy current time into this struct. It is preset to the
 *              current process time.
 * @param priv_data Private data pointer passed when calling pdserv_parameter()
 *
 * \returns 0 on success
 */
typedef int (*write_parameter_t)(
        const struct pdvariable *param,
        void *dst,
        const void *src,
        size_t len,
        struct timespec* time,
        void *priv_data
        );

/** Register a parameter
 *
 * This call registers a parameter. See pdserv_signal() for the description of
 * similar function arguments.
 *
 * During the registration of a parameter, the caller can supply a callback
 * function \p write_cb of type \ref write_parameter_t that is called with
 * data source and destination addresses when a parameter is changed. \p
 * write_cb is responsible to copy the data, usually using memcpy(). If custom
 * locking is used (see pdserv_set_parameter_writelock_cb()), that function
 * will be called before and after \b write_cb.
 *
 * By default memcpy() will be used automatically when \p write_cb is NULL.
 */
PDSERV_EXPORT struct pdvariable *pdserv_parameter(
        struct pdserv* pdserv,    /**< Handle to pdserv */
        const char *path,         /**< Parameter path */
        unsigned int mode,        /**< Access mode, same as Unix file mode */
        int datatype,             /**< Parameter data type identifier:
                                   * - pd_*_T
                                   * - compound identifier from
                                   *   pdserv_create_compound() */
        void *addr,               /**< Parameter address */
        size_t n,                 /**< Parameter element count.
                                   * If \p dim != NULL, this
                                   * is the number elements in \p dim */
        const size_t *dim,        /**< Dimensions. If NULL, the parameter
                                   * is a vector of length \p n */
        write_parameter_t write_cb, /**< Callback for updating the parameter
                                     * inside real time context */
        void *priv_data           /**< Arbitrary pointer passed to \p write_cb
                                   * when it is called */
        );

/** Set an optional alias for a variable
 *
 * Sometimes it it easier to find a variable by its alias. Use this
 * function after calling either pdserv_signal() or pdserv_parameter()
 * to set the alias name for a variable.
 */
PDSERV_EXPORT void pdserv_set_alias(
        struct pdvariable *variable, /**< Parameter or Signal address */
        const char *alias       /**< Variable's alias */
        );

/** Set the optional unit of a variable */
PDSERV_EXPORT void pdserv_set_unit(
        struct pdvariable *variable, /**< Parameter or Signal address */
        const char *unit        /**< Variable's unit */
        );

/** Set the optional comment of a variable */
PDSERV_EXPORT void pdserv_set_comment(
        struct pdvariable *variable, /**< Parameter or Signal address */
        const char *comment     /**< Variable's comment */
        );

/** Get the path of a registered Parameter or Signal.
 *
 * Useful for example in Parameter Update Event Callback.
 * \see write_parameter_t
 *
 * \returns Path of given variable.
*/
PDSERV_EXPORT const char *pdserv_get_variable_path(
        const struct pdvariable *variable /**< Parameter or Signal address */
        );

/** Finish initialisation
 *
 * This function is called after registering all variables and parameters,
 * just before going into cyclic real time. This will fork off the process
 * that communicates with the rest of the world.
 *
 * Since the new process lives in a separate memory space, all parameters must
 * have been initialised beforehand. After calling pdserv_prepare(), parameters
 * should not be changed any more. They can only be updated though a call by
 * the library.
 *
 * \returns 0 on Success, nonzero on Error (usually the value of \c errno ).
 */
PDSERV_EXPORT int pdserv_prepare(
        struct pdserv* pdserv     /**< Handle to pdserv */
        );

/** Update task statistics variables
 *
 * Call this function at the end of the calculation cycle just before
 * pdserv_update() to update statistics
 */
PDSERV_EXPORT void pdserv_update_statistics(
        struct pdtask* pdtask,  /**< Handle to pdtask */
        double exec_time,       /**< Execution time in [s] */
        double cycle_time,      /**< Time since last call */
        unsigned int overrun    /**< total overrun count */
        );

/** Update variables
 *
 * Call this function once at the end of the calculation cycle to update
 * variables
 */
PDSERV_EXPORT void pdserv_update(
        struct pdtask* pdtask,  /**< Handle to pdtask */
        const struct timespec *t  /**< Current model time.
                                  * If NULL, zero is assumed */
        );

/** Cleanup pdserv */
PDSERV_EXPORT void pdserv_exit(
        struct pdserv*           /**< Handle to pdserv */
        );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* PDSERV_H */
